/**
 * <p>Title: Haplotype</p>
 * <p>Description: Approximation Algorithms for Robust tag SNPs</p>
 * <p>Copyright: Copyright (c) 2004</p>
 * <p>Company: CSIE.NTU</p>
 * @author Yao-Ting Huang
 * @version 1.0
 */
import java.util.*;
import java.io.*;
import java.math.BigInteger;

public class Optmize {
  Vector reduction;
  int NumberOfMissingData, size, total=0;
  BufferedWriter bw, out;
  String BlockID;

  public Optmize(int NumberOfMissingData) {
    this.reduction = new Vector();
    this.NumberOfMissingData = NumberOfMissingData;
    this.init();
  }

  private void init(){
   try{
    this.bw = new BufferedWriter(new FileWriter(new File("Reduction_opt.txt")));
    this.out = new BufferedWriter(new FileWriter(new File("Opt_"+this.NumberOfMissingData+".txt")));
   }
   catch(IOException ee){
    ee.printStackTrace();
     }
  }


  public void Reduce(Vector vt, String BlockID) {
    StringBuffer sb = (StringBuffer) vt.elementAt(0);
    this.BlockID = BlockID;

    this.size = vt.size()*(vt.size()-1)/2;

    //reset reduction vector;
    reduction.clear();
    for (int i = 0; i < sb.length(); i++) {
      Bipartite gt = new Bipartite();
      gt.locus = i;
      gt.edges = new int[this.size];
      int index = 0;
      for (int j = 0; j < vt.size(); j++) {
        StringBuffer sb1 = (StringBuffer) vt.elementAt(j);
        for (int k = j + 1; k < vt.size(); k++) {
          StringBuffer sb2 = (StringBuffer) vt.elementAt(k);
          if ((sb1.charAt(i) != sb2.charAt(i)) && sb1.charAt(i)!='n' && sb2.charAt(i)!='n') {
          //if ((sb1.charAt(i) != sb2.charAt(i))) {
            gt.edges[gt.count] = index;
            gt.count++;
          }
          index++;
        }
      }
      //store the transformed result
      reduction.addElement(gt);
    }

    //write reduction result
    try{
      for (int l = 0; l < reduction.size(); l++) {
        Bipartite temp = (Bipartite) reduction.elementAt(l);
        bw.write(BlockID+"\t"+temp.count);
        for(int m=0;m<temp.edges.length;m++){
          bw.write("\t"+temp.edges[m]);
        }
        bw.newLine();
        bw.flush();
      }
    }
    catch (Exception ee) {
      ee.printStackTrace();
    }

  }

  public void run(){
    Integer[] SNP = new Integer[this.reduction.size()];

    if(this.size>0){
      //Initialize the SNP Integer array
      for (int j = 0; j < this.reduction.size(); j++) {
        SNP[j] = new Integer(j);
      }

      //Enumerate each combination of the SNP
      for (int i = 1; i <= this.reduction.size(); i++) {
        //table[] and tablesize are used to represent the uncovered grids in the table
        int[] table;
        int tablesize;
        boolean SolutionFound = false;

	System.out.println(i);

        try {
          Combinations c = new Combinations(SNP, i);
          //Enumerate each (n;i) combination
          while (c.hasMoreElements()) {
            //Extract one combination
            Object[] SNPList = (Object[]) c.nextElement();

            //Initialize the table with m+1 depth
            table = new int[this.size];
            tablesize = this.size * (this.NumberOfMissingData + 1);

            //Set the table with depth m+1
            for (int l = 0; l < table.length; l++) {
              table[l] = this.NumberOfMissingData + 1;
            }

            //Test if this combination of SNPs can cover all grids
            for (int k1 = 0; k1 < SNPList.length; k1++) {
              Bipartite bp = (Bipartite)this.reduction.elementAt( ( (Integer)
                  SNPList[k1]).intValue());
              int count = bp.count;

              while (count > 0) {
                for (int k2 = 0; k2 < bp.count; k2++) {
                  //this column is not finished
                  if (table[bp.edges[k2]] > 0) {
                    table[bp.edges[k2]]--;
                    tablesize--;
                    count--;
                    //this column is finished
                  }
                  else {
                    count--; //this edge is an useless edge
                  }
                } //end of for loop
              } //end of while loop

            } //end of for loop k1

            //Check if all grids are convered by this combination of SNPs
            if (tablesize <= 0) {
              //This combination of SNPs is the optimum solution
              SolutionFound = true;
              //Compute the total number of used tag SNPs
              this.total = this.total + SNPList.length;

              //write this result into the outputfile
              this.out.write(this.BlockID + "\t" + SNPList.length);
              for (int k3 = 0; k3 < SNPList.length; k3++) {
                Bipartite temp = (Bipartite)this.reduction.elementAt( ( (
                    Integer) SNPList[k3]).intValue());
                this.out.write("\t" + temp.locus);
              }

              this.out.newLine();
              this.out.flush();

              System.out.println(this.BlockID);
              //break the while loop
              break;
            }
          } //end of while loop: test other combination of SNPs
        } //end of try

        catch (CombinatoricException e) {
          e.printStackTrace();
        }
        catch (IOException e) {
          e.printStackTrace();
        }

        //if one solution is found, no need to check other combination of SNPs
        //Break the for loop to continue to test the next haplotype block
        if (SolutionFound) {
          break;
        }
      } // end of for loop i
    }
  }//end of run

  public void close(){
   try{
     this.out.write("Total SNPs: "+this.total);
     this.out.flush();
    this.bw.close();
    this.out.close();
   }
   catch(IOException ee){
    ee.printStackTrace();
     }

  }

  /**
   * <p>Title: Haplotype</p>
   *
   * <p>Description: Approximation Algorithms for Robust tag SNPs</p>
   *
   * <p>Copyright: Copyright (c) 2004</p>
   *
   * <p>Company: CSIE.NTU</p>
   * @author Yao-Ting Huang
   * @version 1.0
   */
  private class Bipartite implements Serializable{
    /**
     * Bipartite
     *
     * @param length int
     */

    int count = 0; //number of edges
    int locus;  //locus of the SNP
    int[] edges;
  }
}
